@@ -1,7 +1,7 @@
 /*
 
     YABASIC ---  a simple Basic Interpreter
-    written by Marc Ihm 1995-2016
+    written by Marc Ihm 1995-2017
     more info at www.yabasic.de
 
     function.c --- code for functions
@@ -47,7 +47,7 @@ struct command *lastdata = NULL;	/* used to associate all data-commands with eac
 static struct buff_chain *buffroot;	/* start of sys-input buffer */
 static struct buff_chain **buffcurr;	/* current entry in buff_chain */
 static int buffcount;		/* number of filled buffers */
-
+char *last_inkey;         /* last result of inkey$ */
 
 /* ------------- subroutines ---------------- */
 
@@ -288,7 +288,7 @@ create_changestring (int type)	/* create command 'changestring' */
 {
     struct command *cmd;
 
-    cmd = add_command (cCHANGESTRING, FALSE);
+    cmd = add_command (cCHANGESTRING, FALSE, NULL);
     cmd->args = type;
 }
 
@@ -374,7 +374,7 @@ create_function (int type)	/* create command 'function' */
 {
     struct command *cmd;
 
-    cmd = add_command (cFUNCTION, FALSE);
+    cmd = add_command (cFUNCTION, FALSE, NULL);
     cmd->args = type;
 }
 
@@ -568,6 +568,7 @@ function (struct command *current)	/* performs a function */
         break;
     case fINKEY:
         pointer = inkey (a1->value);
+	strcpy(last_inkey, pointer);
         result = stSTRING;
         break;
     case fAND:
@@ -810,7 +811,7 @@ function (struct command *current)	/* performs a function */
         if (badstream (i, 0)) {
             return;
         }
-        if (!(stream_modes[i] & (smREAD | smWRITE))) {
+        if (!(stream_modes[i] & (stmREAD | stmWRITE))) {
             sprintf (string, "stream %d not opened", i);
             error (ERROR, string);
             value = 0;
@@ -1034,6 +1035,7 @@ getmousexybm (char *s, int *px, int *py, int *pb, int *pm)	/* get mouse coordina
     int x = 0, y = 0, b = 0, m = 0;
     char c;
 
+    if (!*s) s=last_inkey;
     if (*s) {
         sscanf (s, "MB%d%c+%d:%04d,%04d", &b, &c, &m, &x, &y);
         if (px) {
@@ -1347,9 +1349,9 @@ create_poke (char flag)		/* create Command 'cPOKE' */
     struct command *cmd;
 
     if (flag == 'S' || flag == 'D') {
-        cmd = add_command (cPOKEFILE, FALSE);
+        cmd = add_command (cPOKEFILE, FALSE, NULL);
     } else {
-        cmd = add_command (cPOKE, FALSE);
+        cmd = add_command (cPOKE, FALSE, NULL);
     }
     cmd->tag = flag;
 }
@@ -1361,6 +1363,8 @@ poke (struct command *cmd)	/* poke into internals */
     char *dest, *s, c;
     char *sarg = NULL;
     double darg;
+    struct stackentry *stack;
+    int count;
 
     if (cmd->tag == 's') {
         sarg = pop (stSTRING)->pointer;
@@ -1425,10 +1429,22 @@ poke (struct command *cmd)	/* poke into internals */
         fputs (sarg, stdout);
     } else if (!strcmp (dest, "random_seed") && !sarg) {
         srand((unsigned int) darg);
+    } else if (!strcmp (dest, "__assert_stack_size") && !sarg) {
+	count = -1;
+	stack = stackhead;
+	while ( stack != stackroot) {
+	    count++;
+	    stack = stack->prev;
+	}
+	if (count != (int) darg) {
+	    sprintf (string, "assertion failed for number of entries on stack; expected = %d, actual = %d",(int) darg,count);
+	    error (FATAL, string);
+	}
     } else if (dest[0] == '#') {
         error (ERROR, "don't use quotes when poking into file");
     } else {
-        error (ERROR, "invalid poke");
+	sprintf(string,"invalid poke: '%s'",dest);
+        error (ERROR, string);
     }
     return;
 }
@@ -1452,7 +1468,7 @@ pokefile (struct command *cmd)	/* poke into file */
         return;
     }
 
-    if (!(stream_modes[stream] & smWRITE)) {
+    if (!(stream_modes[stream] & stmWRITE)) {
         sprintf (string, "Stream %d not open for writing", stream);
         error (ERROR, string);
         return;
@@ -1511,7 +1527,7 @@ peekfile (int stream)		/* read a byte from stream */
     if (stream && badstream (stream, 0)) {
         return 0;
     }
-    if (stream && !(stream_modes[stream] & smREAD)) {
+    if (stream && !(stream_modes[stream] & stmREAD)) {
         sprintf (string, "stream %d not open for reading", stream);
         error (ERROR, string);
         return 0;
@@ -1598,7 +1614,7 @@ create_exception (int flag)	/* create command 'exception' */
 {
     struct command *cmd;
 
-    cmd = add_command (cEXCEPTION, FALSE);
+    cmd = add_command (cEXCEPTION, FALSE, NULL);
     cmd->args = flag;
 }
 
@@ -1644,7 +1660,7 @@ create_restore (char *label)	/* create command 'restore' */
 {
     struct command *c;
 
-    c = add_command (cRESTORE, FALSE);
+    c = add_command (cRESTORE, FALSE, label);
     c->pointer = my_strdup (label);
 }
 
@@ -1662,7 +1678,7 @@ restore (struct command *cmd)	/* reset data pointer to given label */
             /* no label, restore to first command */
             label = cmd->lib->firstdata;
         } else {
-            label = search_label (cmd->pointer, smLABEL | smGLOBAL);
+            label = search_label (cmd->pointer, srmLABEL | srmGLOBAL);
             if (!label) {
                 /* did not find label */
                 sprintf (string, "can't find label '%s'",
@@ -1692,7 +1708,7 @@ create_dbldata (double value)	/* create command dbldata */
 {
     struct command *c;
 
-    c = add_command (cDATA, FALSE);
+    c = add_command (cDATA, FALSE, NULL);
     c->pointer = my_malloc (sizeof (double));
     if (lastdata) {
         lastdata->nextassoc = c;
@@ -1708,7 +1724,7 @@ create_strdata (char *value)	/* create command strdata */
 {
     struct command *c;
 
-    c = add_command (cDATA, FALSE);
+    c = add_command (cDATA, FALSE, NULL);
     if (lastdata) {
         lastdata->nextassoc = c;
     }
@@ -1723,7 +1739,7 @@ create_readdata (char type)	/* create command readdata */
 {
     struct command *cmd;
 
-    cmd = add_command (cREADDATA, FALSE);
+    cmd = add_command (cREADDATA, FALSE, NULL);
     cmd->tag = type;
 }
 
@@ -1788,7 +1804,7 @@ create_dblrelop (char c)	/* create command dblrelop */
         type = cGE;
         break;
     }
-    add_command (type, FALSE);
+    add_command (type, FALSE, NULL);
 }
 
 
@@ -1851,7 +1867,7 @@ create_strrelop (char c)	/* create command strrelop */
         type = cSTRGE;
         break;
     }
-    add_command (type, FALSE);
+    add_command (type, FALSE, NULL);
 }
 
 
@@ -1894,27 +1910,36 @@ switch_compare (void)		/* compare topmost values for switch statement */
 {
     struct stackentry *result, *first, *second;
     double r = 0.;
-
+    
     first = pop (stANY);
     second = stackhead->prev;
+
+    /* stSWITCH_STRING and stSWITCH_NUMBER compare true to any string or number */
     if ((second->type == stSWITCH_STRING || second->type == stSTRING)
-            && first->type == stSTRING) {
-        if (second->type == stSWITCH_STRING) {
-            r = 1.;
-        } else {
-            r = (strcmp (first->pointer, second->pointer) == 0) ? 1. : 0.;
-        }
+	&& first->type == stSTRING) {
+	if (second->type == stSWITCH_STRING) {
+	    r = 1.;
+	} else {
+	    r = (strcmp (first->pointer, second->pointer) == 0) ? 1. : 0.;
+	}
     } else if ((second->type == stSWITCH_NUMBER || second->type == stNUMBER)
-               && first->type == stNUMBER) {
-        if (second->type == stSWITCH_NUMBER) {
-            r = 1.;
-        } else {
-            r = (first->value == second->value);
-        }
+	       && first->type == stNUMBER) {
+	if (second->type == stSWITCH_NUMBER) {
+	    r = 1.;
+	} else {
+	    r = (first->value == second->value) ? 1. : 0.;
+	}
     } else {
-        error (ERROR,
-               "Cannot mix strings and numbers in a single switch statement");
+	error (ERROR,
+	       "mixing strings and numbers in a single switch statement is not allowed");
     }
+
+    /* if comparison was successful once, remember this for all future comparisons */
+    if (r == 1.) {
+	if (second->type == stNUMBER) second->type=stSWITCH_NUMBER;
+	if (second->type == stSTRING) second->type=stSWITCH_STRING;
+    }
+
     result = push ();
     result->type = stNUMBER;
     result->value = r;
@@ -1956,7 +1981,7 @@ create_boole (char c)		/* create command boole */
         type = cNOT;
         break;
     }
-    add_command (type, FALSE);
+    add_command (type, FALSE, NULL);
 }
 
 
