@@ -1,7 +1,7 @@
 /*
 
     YABASIC ---  a simple Basic Interpreter
-    written by Marc Ihm 1995-2016
+    written by Marc Ihm 1995-2017
     more info at www.yabasic.de
 
     graphic.c --- code for windowed graphics, printing and plotting
@@ -82,7 +82,7 @@ int winwidth, winheight;	/* size of window */
 static int winx, winy;		/* position of window */
 
 /* mouse and keyboard */
-int mousex = 0, mousey = 0, mouseb = 0, mousemod = 0;	/* last know mouse coordinates */
+int mousex = 0, mousey = 0, mouseb = 0, mousemod = 0;	/* last known mouse coordinates */
 char *ykey[kLASTKEY + 1];	/* keys returned by inkey */
 #ifdef UNIX
 #else
@@ -150,7 +150,7 @@ create_openwin (int fnt)
     /* create Command 'openwin' */
     struct command *cmd;
 
-    cmd = add_command (cOPENWIN, FALSE);
+    cmd = add_command (cOPENWIN, NULL, NULL);
     cmd->args = fnt;
 }
 
@@ -883,7 +883,7 @@ change_font (char *fname)
                 XUnloadFont (display, myfont->fid);
             }
         }
-        my_free(fontname);
+        if (fontname!=fname) my_free(fontname);
     }
     fontname = my_strdup (fname);
     myfont = XLoadQueryFont (display, fontname);
@@ -1086,7 +1086,7 @@ create_line (int numpoints)
     /* create Command 'line' */
     struct command *cmd;
 
-    cmd = add_command (cLINE, FALSE);
+    cmd = add_command (cLINE, NULL, NULL);
     cmd->args = numpoints;
     cmd->tag = dmNORMAL;
 }
@@ -2219,7 +2219,7 @@ create_openprinter (int num)
     /* create command 'openprinter' */
     struct command *cmd;
 
-    cmd = add_command (cOPENPRN, FALSE);
+    cmd = add_command (cOPENPRN, NULL, NULL);
     cmd->args = num;
 }
 
