@@ -1,7 +1,7 @@
 /*
 
     YABASIC ---  a simple Basic Interpreter
-    written by Marc Ihm 1995-2016
+    written by Marc Ihm 1995-2017
     more info at www.yabasic.de
 
     io.c --- code for screen and file i/o
@@ -106,7 +106,7 @@ create_print(char type)	/* create command 'print' */
 {
     struct command *cmd;
 
-    cmd = add_command(cPRINT, NULL);
+    cmd = add_command(cPRINT, NULL, NULL);
     cmd->pointer = my_malloc(sizeof (int));
     /* store type of print  */
     cmd->tag = type;
@@ -298,7 +298,7 @@ curinit(void)			/* initialize curses */
 
 
 char *
-inkey(double maxtime)		/* get char from keyboard */
+inkey (double maxtime)		/* get char from keyboard */
 {
     char retkey[100];
     long int max, now, start, left;
@@ -374,7 +374,7 @@ inkey(double maxtime)		/* get char from keyboard */
         /* already got key ? */
 #ifdef UNIX
         if (winopened && display && XPending(display)) {
-            getwinkey(retkey);
+            getwinkey (retkey);
             if (retkey[0]) {
                 break;
             }
@@ -431,7 +431,7 @@ inkey(double maxtime)		/* get char from keyboard */
         if (FD_ISSET(termfd, &readfds)) {
             gettermkey(retkey);
         } else if (FD_ISSET(winfd, &readfds)) {
-            getwinkey(retkey);
+            getwinkey (retkey);
         }
 #elif WINDOWS
         if (wait_result == WAIT_OBJECT_0 + 1) {
@@ -674,7 +674,7 @@ create_myopen(int num)		/* create command 'myopen' */
 {
     struct command *cmd;
 
-    cmd = add_command(cOPEN, NULL);
+    cmd = add_command(cOPEN, NULL, NULL);
     cmd->tag = num;
 }
 
@@ -694,7 +694,7 @@ myopen(struct command *cmd)	/* open specified file for given name */
     char **pmode;
     static char *valid_modes[] = { "r", "w", "a", "rb", "wb", "ab", "" };
     static int smodes[] =
-    { smREAD, smWRITE, smWRITE, smREAD, smWRITE, smWRITE };
+    { stmREAD, stmWRITE, stmWRITE, stmREAD, stmWRITE, stmWRITE };
     int smode;
     struct stackentry *p;
     int has_mode, has_stream, printer;
@@ -720,7 +720,7 @@ myopen(struct command *cmd)	/* open specified file for given name */
     } else {
         stream = 0;
         for (i = 1; i < FOPEN_MAX - 4; i++) {
-            if (stream_modes[i] == smCLOSED) {
+            if (stream_modes[i] == stmCLOSED) {
                 stream = i;
                 break;
             }
@@ -748,7 +748,7 @@ myopen(struct command *cmd)	/* open specified file for given name */
         errorcode = 9;
         goto open_done;
     }
-    if (stream_modes[stream] != smCLOSED) {
+    if (stream_modes[stream] != stmCLOSED) {
         sprintf(errorstring, "stream already in use");
         errorcode = 2;
         goto open_done;
@@ -845,7 +845,7 @@ myclose(void)			/* close the specified stream */
     if (abs(s) == STDIO_STREAM || badstream(s, 0)) {
         return;
     }
-    if (stream_modes[s] == smCLOSED) {
+    if (stream_modes[s] == stmCLOSED) {
         sprintf(string, "stream %d already closed", s);
         error(WARNING, string);
         return;
@@ -865,7 +865,7 @@ myclose(void)			/* close the specified stream */
         fclose(streams[s]);
     }
     streams[s] = NULL;
-    stream_modes[s] = smCLOSED;
+    stream_modes[s] = stmCLOSED;
 }
 
 
@@ -906,7 +906,7 @@ myseek(struct command *cmd)	/* reposition file pointer */
     if (abs(s) == STDIO_STREAM || badstream(s, 0)) {
         return;
     }
-    if (!(stream_modes[s] & (smREAD | smWRITE))) {
+    if (!(stream_modes[s] & (stmREAD | stmWRITE))) {
         sprintf(errorstring, "stream %d not open", s);
         errorcode = 11;
         return;
@@ -922,17 +922,17 @@ myseek(struct command *cmd)	/* reposition file pointer */
 
 
 void
-create_pps(int type, int input)	/* create command pushswitch or popswitch */
+create_pps(int type, int input)	/* create command push_stream or pop_stream */
 {
     struct command *cmd;
 
-    cmd = add_command(type, NULL);
+    cmd = add_command(type, NULL, NULL);
     cmd->args = input;
 }
 
 
 void
-push_switch(struct command *cmd)	/* push current stream on stack and switch to new one */
+push_stream(struct command *cmd)	/* push current stream on stack and switch to new one */
 {
     static int oldstream = STDIO_STREAM;
     struct stackentry *s;
@@ -955,12 +955,12 @@ push_switch(struct command *cmd)	/* push current stream on stack and switch to n
         error(DEBUG, string);
     }
     oldstream = stream;
-    myswitch(stream);
+    mystream(stream);
 }
 
 
 void
-pop_switch(void)		/* pop current stream from stack and switch to it */
+pop_stream(void)		/* pop stream from stack and switch to it */
 {
     int stream;
 
@@ -969,12 +969,12 @@ pop_switch(void)		/* pop current stream from stack and switch to it */
         sprintf(string, "popping %d from stack, switching to it", stream);
         error(DEBUG, string);
     }
-    myswitch(stream);
+    mystream(stream);
 }
 
 
 void
-myswitch(int stream)		/* switch to specified stream */
+mystream(int stream)		/* switch to specified stream */
 {
     int stdio, input;
 
@@ -1011,13 +1011,13 @@ checkstream(void)		/* test if currst is still valid */
     input = (currstr > 0);
 
     if (!stdio) {
-        if (input && !(stream_modes[abs(currstr)] & smREAD)) {
+        if (input && !(stream_modes[abs(currstr)] & stmREAD)) {
             sprintf(string, "stream %d not open for reading",
                     abs(currstr));
             error(ERROR, string);
             return FALSE;
         }
-        if (!input && !(stream_modes[abs(currstr)] & (smWRITE | smPRINT))) {
+        if (!input && !(stream_modes[abs(currstr)] & (stmWRITE | stmPRINT))) {
             sprintf(string, "stream %d not open for writing or printing",
                     abs(currstr));
             error(ERROR, string);
@@ -1040,7 +1040,7 @@ testeof(struct command *cmd)	/* close the specified stream */
     }
     result = push();
     result->type = stNUMBER;
-    if (s && !(stream_modes[s] & smREAD)) {
+    if (s && !(stream_modes[s] & stmREAD)) {
         result->value = 1.;
         return;
     }
@@ -1083,7 +1083,7 @@ create_myread(char type, int tileol)	/* create command 'read' */
 {
     struct command *cmd;
 
-    cmd = add_command(cREAD, NULL);
+    cmd = add_command(cREAD, NULL, NULL);
     cmd->args = tileol;		/* true, if read should go til eol */
     cmd->tag = type;		/* can be 'd' or 's' */
 }
@@ -1281,7 +1281,7 @@ create_onestring(char *str)	/* create command 'onestring' */
 {
     struct command *cmd;
 
-    cmd = add_command(cONESTRING, NULL);
+    cmd = add_command(cONESTRING, NULL, NULL);
     cmd->pointer = my_strdup(str);
 }
 
@@ -1327,7 +1327,7 @@ create_colour(int flag)	/* create command 'colour' */
 {
     struct command *c;
 
-    c = add_command(cCOLOUR, NULL);
+    c = add_command(cCOLOUR, NULL, NULL);
     c->args = flag;
 }
 
