@@ -1,7 +1,7 @@
 /*
 
     YABASIC ---  a simple Basic Interpreter
-    written by Marc Ihm 1995-2016
+    written by Marc Ihm 1995-2017
     more info at www.yabasic.de
 
     main.c --- main() and auxilliary functions
@@ -207,6 +207,8 @@ main (int argc, char **argv)
 #endif
 
     time (&compilation_start);
+    last_inkey=my_malloc(INBUFFLEN);
+    last_inkey[0]='\0';
     error (DEBUG, "this is yabasic " VERSION);
     initialize ();
     program_state = INITIALIZED;
@@ -238,7 +240,7 @@ main (int argc, char **argv)
         create_docu_array ();
     }
 
-    add_command (cEND, NULL);
+    add_command (cEND, NULL, NULL);
     sprintf (string, "read %d line(s) and generated %d command(s)", mylineno,
              commandcount);
     error (NOTE, string);
@@ -292,129 +294,129 @@ main (int argc, char **argv)
 
 
 void
-std_diag (char *head, int type, char *name)	/* produce standard diagnostic */
+std_diag (char *head, int type, char *symname, char *diag)	/* produce standard diagnostic */
 {
     int n, i;
-    char *s;
+    char *dest = string;
     struct stackentry *sp;
 
-    if (infolevel >= DEBUG) {
-        s = string;
-        if (type > cLAST_COMMAND || type < cFIRST_COMMAND) {
-            sprintf (s, "%s Illegal %d %n", head, type, &n);
-        } else {
-            if (name)
-                sprintf (s, "%s '%s' (%s) %n", head, explanation[type],
-                         name ? name : "NULL", &n);
-            else {
-                sprintf (s, "%s '%s' %n", head, explanation[type], &n);
-            }
-        }
-        s += n;
-        if (stackhead->prev != stackroot) {
-            sprintf (s, "t[");
-            s += 2;
-            sp = stackhead;
-            for (i = 0; TRUE; i++) {
-                sp = sp->prev;
-                if (sp == stackroot) {
-                    break;
-                }
-                if (i >= 5) {
-                    continue;
-                }
-                if (i > 0) {
-                    sprintf (s, ",");
-                    s++;
-                }
-                switch (sp->type) {
-                case stGOTO:
-                    sprintf (s, "goto%n", &n);
-                    break;
-                case stSTRINGARRAYREF:
-                case stNUMBERARRAYREF:
-                    if (sp->pointer) {
-                        sprintf (s, "%s()%n", (char *) sp->pointer, &n);
-                    } else {
-                        sprintf (s, "ARRAY()%n", &n);
-                    }
-                    break;
-                case stSTRING:
-                    sprintf (s, "'%s'%n", (char *) sp->pointer, &n);
-                    break;
-                case stNUMBER:
-                    sprintf (s, "%g%n", sp->value, &n);
-                    break;
-                case stLABEL:
-                    sprintf (s, "label%n", &n);
-                    break;
-                case stRETADD:
-                    sprintf (s, "retadd%n", &n);
-                    break;
-                case stRETADDCALL:
-                    sprintf (s, "retaddcall%n", &n);
-                    break;
-                case stSWITCH_MARK:
-                    sprintf (s, "switch_mark%n", &n);
-                    break;
-                case stFREE:
-                    sprintf (s, "free%n", &n);
-                    break;
-                case stROOT:
-                    sprintf (s, "root%n", &n);
-                    break;
-                case stSWITCH_STRING:
-                    sprintf (s, "switch_string%n", &n);
-                    break;
-                case stSWITCH_NUMBER:
-                    sprintf (s, "switch_number%n", &n);
-                    break;
-                default:
-                    sprintf (s, "unknown%n", &n);
-                    break;
-                }
-                s += n;
-            }
-            if (i > 5) {
-                sprintf (s, ";+%d%n", i - 5, &n);
-                s += n;
-            }
-            strcat (s, "]b");
-        }
-        error (DEBUG, string);
+    if (type > cLAST_COMMAND || type < cFIRST_COMMAND) {
+	sprintf (dest, "%s Illegal %d %n", head, type, &n);
+    } else {
+	if (symname)
+	    sprintf (dest, "%s %s (%s) %n", head, explanation[type],
+		     symname, &n);
+	else {
+	    sprintf (dest, "%s %s %n", head, explanation[type], &n);
+	}
+	if (diag && *diag) {
+	    dest += n;
+	    sprintf (dest, "'%s' %n", diag, &n);
+	}
     }
+    dest += n;
+    if (stackhead->prev != stackroot) {
+	sprintf (dest, "t[");
+	dest += 2;
+	sp = stackhead;
+	for (i = 0; TRUE; i++) {
+	    sp = sp->prev;
+	    if (sp == stackroot) {
+		break;
+	    }
+	    if (i >= 5) {
+		continue;
+	    }
+	    if (i > 0) {
+		sprintf (dest, ",");
+		dest++;
+	    }
+	    switch (sp->type) {
+	    case stGOTO:
+		sprintf (dest, "goto%n", &n);
+		break;
+	    case stSTRINGARRAYREF:
+	    case stNUMBERARRAYREF:
+		if (sp->pointer) {
+		    sprintf (dest, "%s()%n", (char *) sp->pointer, &n);
+		} else {
+		    sprintf (dest, "ARRAY()%n", &n);
+		}
+		break;
+	    case stSTRING:
+		sprintf (dest, "'%s'%n", (char *) sp->pointer, &n);
+		break;
+	    case stNUMBER:
+		sprintf (dest, "%g%n", sp->value, &n);
+		break;
+	    case stLABEL:
+		sprintf (dest, "label%n", &n);
+		break;
+	    case stRET_ADDR:
+		sprintf (dest, "retadd%n", &n);
+		break;
+	    case stRET_ADDR_CALL:
+		sprintf (dest, "retaddcall%n", &n);
+		break;
+	    case stFREE:
+		sprintf (dest, "free%n", &n);
+		break;
+	    case stROOT:
+		sprintf (dest, "root%n", &n);
+		break;
+	    case stSWITCH_STRING:
+		sprintf (dest, "switch_string%n", &n);
+		break;
+	    case stSWITCH_NUMBER:
+		sprintf (dest, "switch_number%n", &n);
+		break;
+	    default:
+		sprintf (dest, "unknown%n", &n);
+		break;
+	    }
+	    dest += n;
+	}
+	if (i > 5) {
+	    sprintf (dest, ";+%d%n", i - 5, &n);
+	    dest += n;
+	}
+	strcpy (dest, "]b");
+    } else {
+	strcpy (dest, "t[]b");
+    }
+    error (DEBUG, string);
 }
 
-
 struct command *
-add_command (int type, char *name)
+add_command (int type, char *symname, char *diag)
 /* get room for new command, and make a link from old one */
 {
     struct command *new;
 
     if (infolevel >= DEBUG) {
-        std_diag ("creating", type, name);
+        std_diag ("creating", type, symname, diag);
     }
     cmdhead->type = type;		/* store command */
     cmdhead->line = mylineno;
     cmdhead->lib = currlib;
     cmdhead->cnt = commandcount;
-    if (!name || !*name) {
-        cmdhead->name = NULL;
+    if (!symname || !*symname) {
+        cmdhead->symname = NULL;
     } else {
-        cmdhead->name = my_strdup (name);
+        cmdhead->symname = my_strdup (symname);
     }
     commandcount++;
     cmdhead->pointer = NULL;	/* no data yet */
+    cmdhead->tag = 0;
     cmdhead->jump = NULL;
     cmdhead->nextassoc = NULL;
-    cmdhead->switchtick = switchtick;
+    cmdhead->diag = my_strdup(diag);
     if (!currlib->datapointer && cmdhead->type == cDATA) {
         currlib->firstdata = currlib->datapointer = cmdhead;
     }
 
     /* link into chain of commands referencing a symbol */
-    if (name) {
+    if (symname) {
         if (lastref) {
             lastref->nextref = cmdhead;
         }
@@ -430,7 +432,7 @@ add_command (int type, char *name)
     new->symbol = NULL;
     new->nextref = NULL;
     new->nextassoc = NULL;
-    new->name = NULL;
+    new->symname = NULL;
 
     cmdhead->next = new;
     lastcmd = cmdhead;
@@ -438,6 +440,16 @@ add_command (int type, char *name)
     return lastcmd;
 }
 
+struct command *
+add_command_with_switch_state (int type)	/* same as add_command, but add switch_state too */
+{
+    struct command *cmd;
+
+    cmd = add_command (type, NULL, NULL);
+    add_switch_state(cmd);
+}
+
+
 void
 dump_commands (char *dumpfilename)
 /* dump commands into given file */
@@ -473,7 +485,7 @@ dump_commands (char *dumpfilename)
 
     for (cmd=cmdroot; cmd; cmd=cmd->next) {
         sprintf(string,"");
-        fprintf(dump,"Line %4d, Address %p:   %-*s   (lib %s, stk %d, ptr %p, tag 0x%x%s)\n",cmd->line,cmd,max_explanation_length,explanation[cmd->type],cmd->lib->s,cmd->switchtick,cmd->pointer,cmd->tag,string);
+        fprintf(dump,"Line %4d, Address %p:   %-*s   (lib %s, ptr %p, tag 0x%x%s)\n",cmd->line,cmd,max_explanation_length,explanation[cmd->type],cmd->lib->s,cmd->pointer,cmd->tag,string);
         if (cmd->type==cEND) {
             break;
         }
@@ -884,7 +896,7 @@ end_it (void)			/* perform shutdown-operations */
 #else
     if (!Commandline && endreason != erREQUEST) {
 #endif
-        myswitch (STDIO_STREAM);
+        mystream (STDIO_STREAM);
         onestring ("---Program done, press RETURN---\n");
 #ifdef WINDOWS
         SetConsoleMode (ConsoleInput, InitialConsole & (~ENABLE_ECHO_INPUT));
@@ -997,6 +1009,9 @@ initialize (void)
     cmdhead = cmdroot; /* list of commands */ ;
     commandcount = 0;
 
+    /* initialize switch_id stack */
+    initialize_switch_id_stack();
+    
     /* add internal string variables */
     s = get_sym ("yabos$", sySTRING, amADD_GLOBAL);
     if (s->pointer) {
@@ -1019,23 +1034,23 @@ initialize (void)
     /* file stuff */
     for (i = 1; i <= 9; i++) {
         streams[i] = NULL;
-        stream_modes[i] = smCLOSED;
+        stream_modes[i] = stmCLOSED;
     }
     streams[0] = stdin;
-    stream_modes[0] = smREAD | smWRITE;
+    stream_modes[0] = stmREAD | stmWRITE;
 #ifdef UNIX
     printerfile = NULL;		/* no ps-file yet */
 #endif
 
     /* array with explanation */
     for (i = cFIRST_COMMAND; i <= cLAST_COMMAND; i++) {
-        explanation[i] = "???";
+        explanation[i] = NULL;
     }
     explanation[cFIRST_COMMAND] = "FIRST_COMMAND";
     explanation[cFINDNOP] = "FINDNOP";
     explanation[cEXCEPTION] = "EXCEPTION";
     explanation[cLABEL] = "LABEL";
-    explanation[cSUBLINK] = "cSUBLINK";
+    explanation[cLINK_SUBR] = "cLINK_SUBR";
     explanation[cTOKEN] = "TOKEN";
     explanation[cTOKEN2] = "TOKEN2";
     explanation[cTOKENALT] = "TOKENALT";
@@ -1050,9 +1065,9 @@ initialize (void)
     explanation[cQGOSUB] = "QGOSUB";
     explanation[cCALL] = "CALL";
     explanation[cQCALL] = "QCALL";
-    explanation[cRETURN] = "RETURN";
-    explanation[cRET_FROM_FUN] = "RET_FROM_FUN";
-    explanation[cRETVAL] = "RETVAL";
+    explanation[cRETURN_FROM_GOSUB] = "RETURN_FROM_GOSUB";
+    explanation[cRETURN_FROM_CALL] = "RETURN_FROM_CALL";
+    explanation[cCHECK_RETURN_VALUE] = "CHECK_RETURN_VALUE";
     explanation[cSWAP] = "SWAP";
     explanation[cDECIDE] = "DECIDE";
     explanation[cANDSHORT] = "ANDSHORT";
@@ -1077,7 +1092,7 @@ initialize (void)
     explanation[cPUSHSYMLIST] = "PUSHSYMLIST";
     explanation[cPOPSYMLIST] = "POPSYMLIST";
     explanation[cMAKELOCAL] = "MAKELOCAL";
-    explanation[cNUMPARAM] = "NUMPARAM";
+    explanation[cCOUNT_PARAMS] = "COUNT_PARAMS";
     explanation[cMAKESTATIC] = "MAKESTATIC";
     explanation[cARRAYLINK] = "ARRAYLINK";
     explanation[cPUSHARRAYREF] = "PUSHARRAYREF";
@@ -1103,6 +1118,10 @@ initialize (void)
     explanation[cNE] = "NE";
     explanation[cSTREQ] = "STREQ";
     explanation[cSTRNE] = "STRNE";
+    explanation[cSTRLT] = "STRLT";
+    explanation[cSTRLE] = "STRLE";
+    explanation[cSTRGT] = "STRGT";
+    explanation[cSTRGE] = "STRGE";
     explanation[cPUSHSTRSYM] = "PUSHSTRSYM";
     explanation[cPOPSTRSYM] = "POPSTRSYM";
     explanation[cPUSHSTR] = "PUSHSTR";
@@ -1167,21 +1186,21 @@ initialize (void)
     explanation[cEND_SWITCH_MARK] = "END_SWITCH_MARK";
     explanation[cBEGIN_LOOP_MARK] = "BEGIN_LOOP_MARK";
     explanation[cEND_LOOP_MARK] = "END_LOOP_MARK";
-    explanation[cPUSH_SWITCH_MARK] = "PUSH_SWITCH_MARK";
-    explanation[cCLEAN_SWITCH_MARK] = "CLEAN_SWITCH_MARK";
     explanation[cSWITCH_COMPARE] = "SWITCH_COMPARE";
     explanation[cNEXT_CASE] = "NEXT_CASE";
     explanation[cNEXT_CASE_HERE] = "NEXT_CASE_HERE";
-    explanation[cBREAK] = "BREAK";
+    explanation[cBREAK_MULTI] = "BREAK_MULTI";
     explanation[cCONTINUE] = "CONTINUE";
     explanation[cBREAK_HERE] = "BREAK_HERE";
     explanation[cCONTINUE_HERE] = "CONTINUE_HERE";
+    explanation[cPOP_MULTI] = "POP_MULTI";
+    explanation[cCHKPROMPT] = "CHKPROMPT";
     explanation[cLAST_COMMAND] = "???";
     for (i = cFIRST_COMMAND; i <= cLAST_COMMAND; i++) {
-        if (explanation[i][0] == '?') {
-            sprintf (string, "??? command %d ???", i);
-            explanation[i] = my_strdup (string);
-        }
+	if (!explanation[i]) {
+	    sprintf (string, "command %d has no description (command after %s)", i, explanation[i-1]);
+	    error(ERROR,string);
+	}
     }
 
     ykey[kERR] = "error";
@@ -1291,7 +1310,7 @@ run_it ()
         while (current != cmdhead) {
             if (current->type == cDOCU) {
                 if (infolevel >= DEBUG) {
-                    std_diag ("executing", current->type, current->name);
+                    std_diag ("executing", current->type, current->symname, current->diag);
                 }
                 printf ("%s\n", (char *) current->pointer);
                 l++;
@@ -1301,7 +1320,7 @@ run_it ()
                 }
             } else {
                 if (infolevel >= DEBUG) {
-                    std_diag ("skipping", current->type, current->name);
+                    std_diag ("skipping", current->type, current->symname, current->diag);
                 }
             }
             current = current->next;
@@ -1316,7 +1335,7 @@ run_it ()
     } else {
         while (current != cmdhead && endreason == erNONE) {
             if (infolevel >= DEBUG) {
-                std_diag ("executing", current->type, current->name);
+                std_diag ("executing", current->type, current->symname, current->diag);
             }
             switch (current->type) {
             case cGOTO:
@@ -1339,7 +1358,7 @@ run_it ()
             case cRESETSKIPONCE:
                 resetskiponce (current);
                 DONE;
-            case cBREAK:
+            case cBREAK_MULTI:
                 mybreak (current);
                 DONE;
             case cNEXT_CASE:
@@ -1348,12 +1367,6 @@ run_it ()
             case cSWITCH_COMPARE:
                 switch_compare ();
                 DONE;
-            case cPUSH_SWITCH_MARK:
-                push_switch_mark ();
-                DONE;
-            case cCLEAN_SWITCH_MARK:
-                clean_switch_mark (current);
-                DONE;
             case cCONTINUE:
                 mycontinue (current);
                 DONE;
@@ -1368,7 +1381,7 @@ run_it ()
             case cDATA:
             case cNOP:
             case cUSER_FUNCTION:
-            case cSUBLINK:
+            case cLINK_SUBR:
             case cEND_FUNCTION:
             case cDOCU:
             case cBREAK_HERE:
@@ -1389,12 +1402,12 @@ run_it ()
             case cEXECUTE2:
                 execute (current);
                 DONE;
-            case cRETURN:
-            case cRET_FROM_FUN:
+            case cRETURN_FROM_GOSUB:
+            case cRETURN_FROM_CALL:
                 myreturn (current);
                 DONE;
-            case cRETVAL:
-                retval (current);
+            case cCHECK_RETURN_VALUE:
+                check_return_value (current);
                 DONE;
             case cPUSHDBLSYM:
                 pushdblsym (current);
@@ -1408,6 +1421,9 @@ run_it ()
             case cPOP:
                 pop (stANY);
                 DONE;
+            case cPOP_MULTI:
+                pop_multi (current);
+                DONE;
             case cPOPSTRSYM:
                 popstrsym (current);
                 DONE;
@@ -1432,8 +1448,8 @@ run_it ()
             case cMAKELOCAL:
                 makelocal (current);
                 DONE;
-            case cNUMPARAM:
-                numparam (current);
+            case cCOUNT_PARAMS:
+                count_params (current);
                 DONE;
             case cMAKESTATIC:
                 makestatic (current);
@@ -1499,10 +1515,10 @@ run_it ()
                 myseek (current);
                 DONE;
             case cPUSHSTREAM:
-                push_switch (current);
+                push_stream (current);
                 DONE;
             case cPOPSTREAM:
-                pop_switch ();
+                pop_stream ();
                 DONE;
             case cCHKPROMPT:
                 chkprompt ();
@@ -1899,15 +1915,15 @@ dotify (char *name, int addfun)	/* add library name, if not already present */
 {
     static char buff[200];
     if (!strchr (name, '.')) {
-        strcpy (buff, currlib->s);
-        strcat (buff, ".");
-        strcat (buff, name);
+	strncpy (buff, currlib->s, 200);
+        strncat (buff, ".", 200-1-strlen(buff));
+        strncat (buff, name, 200-1-strlen(buff));
     } else {
-        strcpy (buff, name);
+        strncpy (buff, name, 200);
     }
     if (addfun && !strchr (name, '@')) {
-        strcat (buff, "@");
-        strcat (buff, current_function);
+        strncat (buff, "@", 200-1-strlen(buff));
+        strncat (buff, current_function, 200-1-strlen(buff));
     }
     return buff;
 }
@@ -1924,9 +1940,9 @@ strip (char *name)		/* strip down to minimal name */
     }
     dot = strchr (name, '.');
     if (dot) {
-        strcpy (buff, dot + 1);
+        strncpy (buff, dot + 1, 300);
     } else {
-        strcpy (buff, name);
+        strncpy (buff, name, 300);
     }
     at = strchr (buff, '@');
     if (at) {
@@ -1945,7 +1961,7 @@ do_error (struct command *cmd)	/* issue user defined error */
 
     s = stackhead;
     while (s != stackroot) {
-        if (s->type == stRETADDCALL) {
+        if (s->type == stRET_ADDR_CALL) {
             r = s->pointer;
             cmd->line = r->line;
             cmd->lib = r->lib;
@@ -1962,7 +1978,7 @@ compile ()			/* create s subroutine at runtime */
 {
     open_string (pop (stSTRING)->pointer);
     yyparse ();
-    add_command (cEND, NULL);
+    add_command (cEND, NULL, NULL);
 }
 
 
@@ -1971,7 +1987,7 @@ create_execute (int string)	/* create command 'cEXECUTESUB' */
 {
     struct command *cmd;
 
-    cmd = add_command (string ? cEXECUTE2 : cEXECUTE, NULL);
+    cmd = add_command (string ? cEXECUTE2 : cEXECUTE, NULL, NULL);
     cmd->pointer = my_strdup (dotify ("", FALSE));
 }
 
@@ -2011,7 +2027,7 @@ execute (struct command *cmd)	/* execute a subroutine */
     strcat (fullname, shortname);
     free (st->pointer);
     st->type = stFREE;
-    newcurr = search_label (fullname, smSUB);
+    newcurr = search_label (fullname, srmSUBR);
     if (!newcurr) {
         sprintf (string, "subroutine '%s' not defined", fullname);
         error (ERROR, string);
@@ -2019,8 +2035,8 @@ execute (struct command *cmd)	/* execute a subroutine */
     }
     ret = push ();
     ret->pointer = current;
-    ret->type = stRETADDCALL;
-    reshufflestack (ret);
+    ret->type = stRET_ADDR_CALL;
+    reorder_stack_before_call (ret);
     current = newcurr;
     free (fullname);
 }
@@ -2035,7 +2051,7 @@ create_docu (char *doc)		/* create command 'docu' */
     if (inlib) {
         return;
     }
-    cmd = add_command (cDOCU, NULL);
+    cmd = add_command (cDOCU, NULL, NULL);
     cmd->pointer = doc;
     if (previous) {
         previous->nextassoc = cmd;
