@@ -1,10 +1,10 @@
 /*
 
     YABASIC ---  a simple Basic Interpreter
-    written by Marc Ihm 1995-2016
+    written by Marc Ihm 1995-2017
     more info at www.yabasic.de
 
-    symbol.c --- code for control structures, symbol and stack management
+    symbol.c --- code for symbol, stack and library management, handling of arrays
 
     This file is part of yabasic and may be copied under the terms of
     MIT License which can be found in the file LICENSE.
@@ -22,13 +22,12 @@
 
 extern int mylineno;		/* current line number */
 extern int yyparse ();		/* call bison parser */
-
+extern int switch_nesting;
+extern int switch_id;
 
 /* ------------- local functions ---------------- */
 
 static struct symbol *create_symbol (int, char *);	/* create a new symbol */
-static void link_label (struct command *);	/* link label into list of labels */
-static int count_args (int);	/* count number of arguments on stack */
 static void stackdesc (int, char *);	/* give back string describing stackentry */
 static void freesym (struct symbol *);	/* free contents of symbol */
 static int ind_to_off (int *, int *);	/* convert array of indices to single offset */
@@ -41,13 +40,10 @@ static struct symstack *symroot = NULL;	/* first element in symbol list */
 static struct symstack *symhead = NULL;	/* last element ind symbol list */
 struct stackentry *stackroot;	/* lowest element in stack */
 struct stackentry *stackhead;	/* topmost element in stack */
-static struct command *labelroot = NULL;	/* first label among commands */
-static struct command *labelhead = NULL;	/* last label seen so far */
 extern char *current_function;	/* name of currently defined function */
 struct command *lastref;	/* last command in UDS referencing a symbol */
 struct command *firstref;	/* first command in UDS referencing a symbol */
 int labelcount = 0;		/* count self-generated labels */
-int switchtick = 0;             /* change in and out of switch-statements *
 
 
 /* ------------- subroutines ---------------- */
@@ -263,69 +259,6 @@ link_symbols (struct symbol *from, struct symbol *to)
 }
 
 
-void
-create_retval (int is, int should)	/* create command 'cRETVAL' */
-{
-    struct command *cmd;
-
-    cmd = add_command (cRETVAL, NULL);
-    cmd->args = is;
-    cmd->tag = should;
-}
-
-
-void
-retval (struct command *cmd)	/* check return value of function */
-{
-    int is, should;
-    struct stackentry *s;
-
-    is = cmd->args;
-    should = cmd->tag;
-    if (is == should) {
-        /* okay, function returns expected type */
-    } else if (is == ftNONE) {
-        /* no element on stack, create one */
-        s = push ();
-        if (should == ftNUMBER) {
-            s->type = stNUMBER;
-            s->value = 0.0;
-        } else {
-            s->type = stSTRING;
-            s->pointer = my_strdup ("");
-        }
-    } else {
-        sprintf (string, "subroutine returns %s but should return %s",
-                 (is == ftSTRING) ? "a string" : "a number",
-                 (should == ftSTRING) ? "a string" : "a number");
-        error (ERROR, string);
-    }
-    if (infolevel >= DEBUG) {
-        s = stackhead->prev;
-        if (s->type == stNUMBER) {
-            sprintf (string, "subroutine returns number %g", s->value);
-        } else if (s->type == stSTRING)
-            sprintf (string, "subroutine returns string '%s'",
-                     (char *) s->pointer);
-        else
-            sprintf (string, "subroutine returns something strange (%d)",
-                     s->type);
-        error (DEBUG, string);
-    }
-    swap ();
-}
-
-
-void
-create_endfunction (void)	/* create command cEND_FUNCTION */
-{
-    struct command *cmd;
-
-    cmd = add_command (cEND_FUNCTION, NULL);
-    link_label (cmd);
-}
-
-
 void
 dump_sym (void)			/* dump the stack of lists of symbols */
 {
@@ -369,43 +302,6 @@ dump_sym (void)			/* dump the stack of lists of symbols */
 }
 
 
-void
-dump_sub (int short_dump)	/* dump the stack of subroutine calls */
-{
-    struct stackentry *st = stackhead;
-    struct command *cmd;
-    int first = TRUE;
-    do {
-        if (st->type == stRETADDCALL) {
-            cmd = st->pointer;
-            if (cmd->type == cCALL || cmd->type == cQCALL) {
-                char *dot;
-                dot = strchr (cmd->pointer, '.');
-                if (first && !short_dump) {
-                    error (DUMP, "Executing in:");
-                }
-                sprintf (string, "sub %s() called in %s,%d",
-                         dot ? (dot + 1) : (char *) cmd->pointer, cmd->lib->l,
-                         cmd->line);
-                error (DUMP, string);
-                first = FALSE;
-            }
-        }
-        st = st->prev;
-    } while (st && st != stackroot);
-    if (first && !short_dump) {
-        if (!short_dump) {
-            error (DUMP, "Executing in:");
-        }
-    }
-    if (!short_dump) {
-        error (DUMP, "main program");
-    }
-
-    return;
-}
-
-
 static struct symbol *
 create_symbol (int type, char *name)  	/* create a new symbol */
 {
@@ -424,30 +320,6 @@ create_symbol (int type, char *name)  	/* create a new symbol */
 }
 
 
-void
-function_or_array (struct command *cmd)	/* decide whether to perform function or array */
-{
-    struct command *fu;
-
-    fu = search_label (cmd->name, smSUB | smLINK);
-    if (fu) {
-        cmd->type = cCALL;
-        cmd->pointer = cmd->name;
-        cmd->name = NULL;
-        error (DEBUG, "converting FUNCTION_OR_ARRAY to FUNCTION");
-    } else {
-        if (cmd->type == cFUNCTION_OR_ARRAY) {
-            cmd->tag = CALLARRAY;
-        } else {
-            cmd->tag = CALLSTRINGARRAY;
-        }
-        cmd->type = cDOARRAY;
-        cmd->args = -1;
-        error (DEBUG, "converting FUNCTION_OR_ARRAY to ARRAY");
-    }
-}
-
-
 void
 swap ()				/* swap topmost elements on stack */
 {
@@ -507,24 +379,23 @@ pop (int etype)
     struct stackentry *s;
 
     /* test if there is something on the stack */
-    if (stackhead == stackroot) {
-        error (FATAL, "Popped too much.");
-        return stackhead;
-    }
+    if (stackhead == stackroot) error (FATAL, "Popped too much.");
     stackhead = stackhead->prev;	/* move down in stack */
     ftype = stackhead->type;
     if (etype == ftype || etype == stANY ||
             (etype == stSTRING_OR_NUMBER
-             && (ftype == stNUMBER || ftype == stSTRING))
+             && (ftype == stNUMBER || ftype == stSWITCH_NUMBER || ftype == stSTRING || ftype == stSWITCH_STRING))
             || (etype == stSTRING_OR_NUMBER_ARRAYREF
-                && (ftype == stSTRINGARRAYREF || ftype == stNUMBERARRAYREF))) {
+                && (ftype == stSTRINGARRAYREF || ftype == stNUMBERARRAYREF))
+	|| (etype == stSTRING && ftype == stSWITCH_STRING)
+	|| (etype == stNUMBER && ftype == stSWITCH_NUMBER)) {
         return stackhead;    /* this is your value; use it quickly ! */
     }
 
     /* expected and found don't match */
     stackdesc (etype, expected);
     stackdesc (ftype, found);
-    sprintf (string, "expected %s but found %s", expected, found);
+    sprintf (string, "expected '%s' but found '%s'", expected, found);
     if (etype == stNUMBER || etype == stSTRING
             || etype == stSTRING_OR_NUMBER) {
         s = push ();
@@ -566,10 +437,10 @@ stackdesc (int type, char *desc)	/* give back string describing stackentry */
     case stLABEL:
         strcpy (desc, "a label");
         break;
-    case stRETADD:
+    case stRET_ADDR:
         strcpy (desc, "a return address for gosub");
         break;
-    case stRETADDCALL:
+    case stRET_ADDR_CALL:
         strcpy (desc, "a return address for a subroutine");
         break;
     case stFREE:
@@ -588,10 +459,10 @@ stackdesc (int type, char *desc)	/* give back string describing stackentry */
         strcpy (desc, "reference to a string or an array");
         break;
     case stSWITCH_STRING:
-        strcpy (desc, "number for switch");
+        strcpy (desc, "string for switch");
         break;
     case stSWITCH_NUMBER:
-        strcpy (desc, "string for switch");
+        strcpy (desc, "number for switch");
         break;
     default:
         sprintf (desc, "type %d", type);
@@ -684,7 +555,7 @@ create_pushdbl (double value)	/* create command 'cPUSHDBL' */
 {
     struct command *cmd;
 
-    cmd = add_command (cPUSHDBL, NULL);
+    cmd = add_command (cPUSHDBL, NULL, NULL);
     cmd->pointer = my_malloc (sizeof (double));
     *(double *) (cmd->pointer) = value;
 }
@@ -709,13 +580,17 @@ pushdblsym (struct command *cmd)
     struct stackentry *p;
 
     p = push ();
-    if (!cmd->name) {
+    if (!cmd->symname) {
         error (WARNING, "invalid pushdblsym");
     }
 
     if (!cmd->symbol) {
-        cmd->symbol = &(get_sym (cmd->name, syNUMBER, amADD_GLOBAL)->value);
+        cmd->symbol = &(get_sym (cmd->symname, syNUMBER, amADD_GLOBAL)->value);
+    } else if (infolevel >= DEBUG) {
+    	sprintf(string, "reading symbol '%s'", cmd->symname);
+	error (DEBUG, string);
     }
+
     p->value = *(double *) cmd->symbol;
     p->type = stNUMBER;
 }
@@ -728,63 +603,22 @@ popdblsym (struct command *cmd)	/* pop double from stack */
 
     d = pop (stNUMBER)->value;
     if (!cmd->symbol) {
-        cmd->symbol = &(get_sym (cmd->name, syNUMBER, amADD_GLOBAL)->value);
+        cmd->symbol = &(get_sym (cmd->symname, syNUMBER, amADD_GLOBAL)->value);
+    } else if (infolevel >= DEBUG) {
+	sprintf(string, "writing symbol '%s'", cmd->symname);
+	error (DEBUG, string);
     }
+	
     *(double *) (cmd->symbol) = d;
 }
 
 
-void
-create_makelocal (char *name, int type)	/* create command 'cMAKELOCAL' */
-{
-    struct command *cmd;
-
-    cmd = add_command (cMAKELOCAL, name);
-    cmd->args = type;
-}
-
-
-void
-makelocal (struct command *cmd)	/* makes symbol local */
-{
-    if (get_sym (cmd->name, cmd->args, amSEARCH_VERY_LOCAL)) {
-        sprintf (string,
-                 "local variable '%s' already defined within this subroutine",
-                 strip (cmd->name));
-        error (ERROR, string);
-        return;
-    }
-    get_sym (cmd->name, cmd->args, amADD_LOCAL);
-}
-
-
-void
-create_numparam (void)		/* create command 'cNUMPARAM' */
-{
-    struct command *cmd;
-
-    /* dotifying numparams at compiletime (as opposed to runtime) is essential,
-       because the function name is not known at runtime */
-    cmd = add_command (cNUMPARAM, dotify ("numparams", FALSE));
-}
-
-
-void
-numparam (struct command *cmd)	/* count number of function parameters */
-{
-    struct symbol *sym;
-
-    sym = get_sym (cmd->name, syNUMBER, amADD_LOCAL);
-    sym->value = abs (count_args (FALSE));
-}
-
-
 void
 create_makestatic (char *name, int type)	/* create command 'cMAKESTATIC' */
 {
     struct command *cmd;
 
-    cmd = add_command (cMAKESTATIC, name);
+    cmd = add_command (cMAKESTATIC, name, NULL);
     cmd->args = type;
 }
 
@@ -797,14 +631,14 @@ makestatic (struct command *cmd)	/* makes symbol static */
 
 
     /* mask function name */
-    if ((at = strchr (cmd->name, '@')) != NULL) {
+    if ((at = strchr (cmd->symname, '@')) != NULL) {
         *at = '\0';
     }
 
-    if (get_sym (cmd->name, cmd->args, amSEARCH_VERY_LOCAL)) {
+    if (get_sym (cmd->symname, cmd->args, amSEARCH_VERY_LOCAL)) {
         sprintf (string,
                  "static variable '%s' already defined within this subroutine",
-                 strip (cmd->name));
+                 strip (cmd->symname));
         error (ERROR, string);
         return;
     }
@@ -813,13 +647,13 @@ makestatic (struct command *cmd)	/* makes symbol static */
     if (at) {
         *at = '@';
     }
-    g = get_sym (cmd->name, cmd->args, amADD_GLOBAL);
+    g = get_sym (cmd->symname, cmd->args, amADD_GLOBAL);
     if (at) {
         *at = '\0';
     }
 
     /* create local variable */
-    l = get_sym (cmd->name, cmd->args, amADD_LOCAL);
+    l = get_sym (cmd->symname, cmd->args, amADD_LOCAL);
     if (at) {
         *at = '@';
     }
@@ -833,7 +667,7 @@ create_arraylink (char *name, int type)	/* create command 'cARRAYLINK' */
 {
     struct command *cmd;
 
-    cmd = add_command (cARRAYLINK, name);
+    cmd = add_command (cARRAYLINK, name, NULL);
     cmd->pointer = current_function;
     cmd->args = type;
 }
@@ -845,16 +679,16 @@ arraylink (struct command *cmd)	/* link a local symbol to a global array */
     struct symbol *l, *g;
     struct array *ar;
 
-    if (get_sym (cmd->name, cmd->args, amSEARCH_VERY_LOCAL)) {
+    if (get_sym (cmd->symname, cmd->args, amSEARCH_VERY_LOCAL)) {
         sprintf (string, "'%s()' already defined within this subroutine",
-                 strip (cmd->name));
+                 strip (cmd->symname));
         error (ERROR, string);
         return;
     }
     /* get globally defined array */
     g = get_sym (pop (cmd->args)->pointer, syARRAY, amSEARCH_PRE);
     /* create local array */
-    l = get_sym (cmd->name, syARRAY, amADD_LOCAL);
+    l = get_sym (cmd->symname, syARRAY, amADD_LOCAL);
     if (!l) {
         return;
     }
@@ -865,7 +699,7 @@ arraylink (struct command *cmd)	/* link a local symbol to a global array */
         l->pointer = ar;
         if (infolevel >= DEBUG) {
             sprintf (string, "creating 0-dimensional dummy array '%s()'",
-                     cmd->name);
+                     cmd->symname);
             error (DEBUG, string);
         }
     } else {
@@ -880,7 +714,7 @@ create_pusharrayref (char *name, int type)	/* create command 'cPUSHARRAYREF' */
 {
     struct command *cmd;
 
-    cmd = add_command (cPUSHARRAYREF, name);
+    cmd = add_command (cPUSHARRAYREF, name, NULL);
     cmd->args = type;
 }
 
@@ -891,7 +725,7 @@ pusharrayref (struct command *cmd)	/* push an array reference onto stack */
     struct stackentry *s;
     s = push ();
     s->type = cmd->args;
-    s->pointer = my_strdup (cmd->name);
+    s->pointer = my_strdup (cmd->symname);
 }
 
 
@@ -900,7 +734,7 @@ create_require (int type)	/* create command 'cREQUIRE' */
 {
     struct command *cmd;
 
-    cmd = add_command (cREQUIRE, NULL);
+    cmd = add_command (cREQUIRE, NULL, NULL);
     cmd->args = type;
 }
 
@@ -974,19 +808,19 @@ create_dblbin (char c)		/* create command for binary double operation */
 {
     switch (c) {
     case '+':
-        add_command (cDBLADD, NULL);
+        add_command (cDBLADD, NULL, NULL);
         break;
     case '-':
-        add_command (cDBLMIN, NULL);
+        add_command (cDBLMIN, NULL, NULL);
         break;
     case '*':
-        add_command (cDBLMUL, NULL);
+        add_command (cDBLMUL, NULL, NULL);
         break;
     case '/':
-        add_command (cDBLDIV, NULL);
+        add_command (cDBLDIV, NULL, NULL);
         break;
     case '^':
-        add_command (cDBLPOW, NULL);
+        add_command (cDBLPOW, NULL, NULL);
         break;
     }
     /* no specific information needed */
@@ -1049,7 +883,7 @@ pushstrptr (struct command *cmd)	/* push string-pointer onto stack */
 
     p = push ();
     if (!cmd->symbol) {
-        cmd->symbol = &(get_sym (cmd->name, sySTRING, amADD_GLOBAL)->pointer);
+        cmd->symbol = &(get_sym (cmd->symname, sySTRING, amADD_GLOBAL)->pointer);
     }
     p->pointer = *(char **) cmd->symbol;
     if (!p->pointer) {
@@ -1066,7 +900,7 @@ pushstrsym (struct command *cmd)	/* push string-symbol onto stack */
 
     p = push ();
     if (!cmd->symbol) {
-        cmd->symbol = &(get_sym (cmd->name, sySTRING, amADD_GLOBAL)->pointer);
+        cmd->symbol = &(get_sym (cmd->symname, sySTRING, amADD_GLOBAL)->pointer);
     }
     p->pointer = my_strdup (*(char **) cmd->symbol);
     p->type = stSTRING;
@@ -1076,11 +910,11 @@ pushstrsym (struct command *cmd)	/* push string-symbol onto stack */
 void
 popstrsym (struct command *cmd)	/* pop string from stack */
 {
-    if (!cmd->name) {
+    if (!cmd->symname) {
         return;
     }
     if (!cmd->symbol) {
-        cmd->symbol = &(get_sym (cmd->name, sySTRING, amADD_GLOBAL)->pointer);
+        cmd->symbol = &(get_sym (cmd->symname, sySTRING, amADD_GLOBAL)->pointer);
     }
     if (*(char **) cmd->symbol != NULL) {
         my_free (*(char **) cmd->symbol);
@@ -1094,7 +928,7 @@ create_pushstr (char *s)	/* creates command pushstr */
 {
     struct command *cmd;
 
-    cmd = add_command (cPUSHSTR, NULL);
+    cmd = add_command (cPUSHSTR, NULL, s);
     cmd->pointer = my_strdup (s);	/* store string */
 }
 
@@ -1112,7 +946,7 @@ pushstr (struct command *cmd)
 
 
 void
-duplicate (void)		/* duplicate topmost element of stack */
+duplicate (void)		/* duplicate topmost number on stack */
 {
     struct stackentry *s;
     double actual;
@@ -1124,298 +958,13 @@ duplicate (void)		/* duplicate topmost element of stack */
 }
 
 
-void
-create_goto (char *label)	/* creates command goto */
-{
-    struct command *cmd;
-
-    cmd = add_command (cGOTO, NULL);
-    /* specific info */
-    cmd->pointer = my_strdup (label);
-}
-
-
-void
-create_gosub (char *label)	/* creates command gosub */
-{
-    struct command *cmd;
-
-    cmd = add_command (cGOSUB, NULL);
-    /* specific info */
-    cmd->pointer = my_strdup (label);
-}
-
-
-void
-create_call (char *label)	/* creates command function call */
-{
-    struct command *cmd;
-
-    cmd = add_command (cCALL, NULL);
-    /* specific info */
-    cmd->pointer = my_strdup (label);
-}
-
-
-static void
-link_label (struct command *cmd)	/* link label into list of labels */
-{
-    if (!labelroot) {
-        labelroot = cmd;
-    } else {
-        labelhead->nextassoc = cmd;
-    }
-    labelhead = cmd;
-}
-
-
-struct command *
-search_label (char *name, int type)  	/* search label */
-{
-    struct command *curr;
-    char *at = NULL;
-
-    curr = labelroot;
-    if (type & smGLOBAL) {
-        at = strchr (name, '@');
-        if (at) {
-            *at = '\0';
-        }
-    }
-    while (curr) {
-        if ((type & smSUB) && curr->type == cUSER_FUNCTION
-                && !strcmp (curr->pointer, name)) {
-            if (at) {
-                *at = '@';
-            }
-            return curr;
-        }
-        if ((type & smLINK) && curr->type == cSUBLINK
-                && !strcmp (curr->pointer, name)) {
-            if (at) {
-                *at = '@';
-            }
-            return curr->next;
-        }
-        if ((type & smLABEL) && curr->type == cLABEL
-                && !strcmp (curr->pointer, name)) {
-            if (at) {
-                *at = '@';
-            }
-            return curr;
-        }
-        curr = curr->nextassoc;
-    }
-    return NULL;
-}
-
-
-void
-jump (struct command *cmd)
-/* jump to specific Label; used as goto, gosub or function call */
-{
-    struct command *label;
-    struct stackentry *ret;
-    int type;
-    char *dot;
-
-    type = cmd->type;
-    if (type == cGOSUB || type == cQGOSUB || type == cCALL || type == cQCALL) {
-        /* leave return address for return */
-        ret = push ();
-        ret->pointer = current;
-        if (type == cGOSUB || type == cQGOSUB) {
-            ret->type = stRETADD;
-        } else {
-            ret->type = stRETADDCALL;
-            reshufflestack (ret);
-        }
-    }
-
-    if (type == cQGOSUB || type == cQGOTO || type == cQCALL) {
-        current = (struct command *) cmd->jump;	/* use remembered address */
-        return;
-    }
-    label = search_label (cmd->pointer, smSUB | smLINK | smLABEL);
-    if (!label && type == cCALL && (dot = strchr (cmd->pointer, '.'))) {
-        strcpy (string, "main");
-        strcat (string, dot);
-        label = search_label (string, smLINK);
-    }
-    if (label) {
-        /* found right label */
-        current = label;		/* jump to new location */
-        /* use the address instead of the name next time */
-        cmd->jump = label;
-        switch (cmd->type) {
-        case cGOTO:
-            cmd->type = cQGOTO;
-            break;
-        case cGOSUB:
-            cmd->type = cQGOSUB;
-            break;
-        case cCALL:
-            cmd->type = cQCALL;
-            break;
-        }
-    } else {
-        /* label not found */
-        sprintf (string, "can't find %s '%s'",
-                 (type == cCALL) ? "subroutine" : "label",
-                 strip ((char *) cmd->pointer));
-        if (strchr (cmd->pointer, '@')) {
-            strcat (string, " (not in this sub)");
-        }
-        error (ERROR, string);
-    }
-
-    /* check, if goto enters or leaves a switch_statement */
-    if (cmd->type == cQGOTO) {
-        if (label->switchtick && !cmd->switchtick) {
-            error (ERROR, "cannot jump into switch-statement");
-        } else if (!label->switchtick && cmd->switchtick) {
-            error (ERROR, "cannot jump out of switch-statement");
-        } else if (label->switchtick != cmd->switchtick) {
-            error (ERROR, "cannot jump between switch statements");
-        }
-    }
-}
-
-
-void
-reshufflestack (struct stackentry *ret)	/* reorganize stack for function call */
-{
-    struct stackentry *a, *b, *c;
-    struct stackentry *top, *bot;
-    struct stackentry *ttop, *bbot;
-    int args;
-
-
-    /* this is a function call; revert stack and shuffle return address to bottom */
-    /* push address below parameters */
-    args = 0;
-    top = a = ret->prev;
-    while (a->type != stFREE) {
-        a = a->prev;
-        args++;
-    }
-    bot = a->next;
-    b = a->prev;
-    /* remove ret */
-    ret->prev->next = ret->next;
-    ret->next->prev = ret->prev;
-    /* squeeze ret between a and b */
-    ret->next = a;
-    a->prev = ret;
-    b->next = ret;
-    ret->prev = b;
-    /* revert stack between top and bot */
-    if (args > 1) {
-        a = bot;
-        b = a->next;
-        bbot = bot->prev;
-        ttop = top->next;
-        for (; args > 1; args--) {
-            a->prev = b;
-            c = b->next;
-            b->next = a;
-            a = b;
-            b = c;
-        }
-        bot->next = ttop;
-        bot->next->prev = bot;
-        top->prev = bbot;
-        top->prev->next = top;
-    }
-}
-
-
-void
-myreturn (struct command *cmd)	/* return from gosub of function call */
-{
-    struct stackentry *address;
-
-    address = pop (stANY);
-    if (cmd->type == cRET_FROM_FUN) {
-        if (address->type != stRETADDCALL) {
-            error (ERROR, "RETURN from a subroutine without CALL");
-            return;
-        }
-    } else {
-        if (address->type != stRETADD) {
-            error (ERROR, "RETURN without GOSUB");
-            return;
-        }
-    }
-    current = (struct command *) address->pointer;
-    return;
-}
-
-
-void
-create_label (char *label, int type)	/* creates command label */
-{
-    struct command *cmd;
-
-    /* check, if label is duplicate */
-    if (search_label (label, smSUB | smLINK | smLABEL)) {
-        sprintf (string, "duplicate %s '%s'",
-                 (type == cLABEL) ? "label" : "subroutine", strip (label));
-        error (ERROR, string);
-        return;
-    }
-
-    cmd = add_command (type, NULL);
-    /* store label */
-    cmd->pointer = my_strdup (label);
-    link_label (cmd);
-}
-
-
-void
-create_sublink (char *label)	/* create link to subroutine */
-{
-    char global[200];
-    char *dot;
-    struct command *cmd;
-
-    if (!inlib) {
-        return;
-    }
-    dot = strchr (label, '.');
-    strcpy (global, "main");
-    strcat (global, dot);
-
-    /* check, if label is duplicate */
-    if (search_label (global, smSUB | smLINK | smLABEL)) {
-        sprintf (string, "duplicate subroutine '%s'", strip (global));
-        error (ERROR, string);
-        return;
-    }
-
-    cmd = add_command (cSUBLINK, NULL);
-    /* store label */
-    cmd->pointer = my_strdup (global);
-    link_label (cmd);
-}
-
-
-void
-decide ()			/*  skips next command, if not 0 on stack */
-{
-    if (pop (stNUMBER)->value != 0) {
-        current = current->next;    /* skip one command */
-    }
-}
-
-
 void
 create_dim (char *name, char type)	/* create command 'dim' */
 /* type can be 's'=string or 'd'=double Array */
 {
     struct command *cmd;
 
-    cmd = add_command (cDIM, name);
+    cmd = add_command (cDIM, name, name);
     cmd->tag = type;		/* type: string or double */
     cmd->args = -1;
 }
@@ -1439,10 +988,10 @@ dim (struct command *cmd)	/* get room for array */
         error (ERROR, "only numerical indices allowed for arrays");
         return;
     }
-    s = get_sym (cmd->name, syARRAY, local ? amADD_LOCAL : amADD_GLOBAL);
-    if (search_label (cmd->name, smSUB | smLINK)) {
+    s = get_sym (cmd->symname, syARRAY, local ? amADD_LOCAL : amADD_GLOBAL);
+    if (search_label (cmd->symname, srmSUBR | srmLINK)) {
         sprintf (string, "array '%s()' conflicts with user subroutine",
-                 strip (cmd->name));
+                 strip (cmd->symname));
         error (ERROR, string);
         return;
     }
@@ -1458,7 +1007,7 @@ dim (struct command *cmd)	/* get room for array */
         if (cmd->args != oar->dimension) {
             sprintf (string,
                      "cannot change dimension of '%s()' from %d to %d",
-                     strip (cmd->name), oar->dimension, cmd->args);
+                     strip (cmd->symname), oar->dimension, cmd->args);
             error (ERROR, string);
         }
     }
@@ -1582,7 +1131,7 @@ off_to_ind (int off, int *bound, int *ind)	/* convert a single offset to an arra
 
 
 void
-query_array (struct command *cmd)	/* query array */
+query_array (struct command *cmd)	/* get value from array */
 {
     int index;
     struct stackentry *s;
@@ -1631,7 +1180,7 @@ create_doarray (char *symbol, int command)	/* creates array-commands */
 {
     struct command *cmd;
 
-    cmd = add_command (cDOARRAY, symbol);
+    cmd = add_command (cDOARRAY, symbol, symbol);
     cmd->tag = command;		/* operation to perform */
     cmd->args = -1;
 }
@@ -1650,10 +1199,10 @@ doarray (struct command *cmd)	/* call an array */
 
 
     if (!cmd->symbol) {
-        sym = get_sym (cmd->name, syARRAY, amSEARCH);
+        sym = get_sym (cmd->symname, syARRAY, amSEARCH);
         if (!sym || !sym->pointer) {
             sprintf (string, "'%s()' is neither array nor subroutine",
-                     strip (cmd->name));
+                     strip (cmd->symname));
             error (ERROR, string);
             return;
         }
@@ -1678,7 +1227,7 @@ doarray (struct command *cmd)	/* call an array */
         } else {
             stack->type = stSTRINGARRAYREF;
         }
-        stack->pointer = my_strdup (cmd->name);
+        stack->pointer = my_strdup (cmd->symname);
         return;
     }
 
@@ -1686,13 +1235,13 @@ doarray (struct command *cmd)	/* call an array */
 
     if (!ar->dimension) {
         sprintf (string, "array parameter '%s()' has not been supplied",
-                 strip (cmd->name));
+                 strip (cmd->symname));
         error (ERROR, string);
         return;
     }
     if (cmd->args != ar->dimension) {
         sprintf (string, "%d indices supplied, %d expected for '%s()'",
-                 cmd->args, ar->dimension, strip (cmd->name));
+                 cmd->args, ar->dimension, strip (cmd->symname));
         error (ERROR, string);
         return;
     }
@@ -1770,7 +1319,7 @@ create_array (int type, int dimension)  	/* create an array */
 }
 
 
-static int
+int
 count_args (int skipfirst)	/* count number of numeric arguments on stack */
 {
     int i = 0;
@@ -1795,264 +1344,3 @@ count_args (int skipfirst)	/* count number of numeric arguments on stack */
 }
 
 
-void
-skipper ()
-/* used for on_goto/gosub, skip specified number of commands */
-{
-    int i, len;
-    struct command *ahead;	/* command to follow */
-
-    len = (int) pop (stNUMBER)->value;
-    i = 1;
-    current = current->next;	/* advance to first goto/gosub */
-    for (i = 1; i < len; i++) {
-        ahead = current->next->next;	/* skip interleaving findnop statement */
-        if (ahead->type == cNOP) {
-            break;
-        } else {
-            current = ahead;
-        }
-    }
-}
-
-
-void
-skiponce (struct command *cmd)	/* skip next command exectly once */
-{
-    if (cmd->tag) {
-        current = current->next;
-    }
-    cmd->tag = 0;
-}
-
-
-void
-resetskiponce (struct command *cmd)	/* find and reset next skip */
-{
-    struct command *c;
-
-    c = cmd;
-    while (c->type != cSKIPONCE) {
-        c = c->next;
-    }
-    c->tag = 1;
-}
-
-void
-push_switch_mark (void)		/* push a switch mark */
-{
-    push ()->type = stSWITCH_MARK;
-}
-
-
-void
-create_clean_switch_mark (int keep, int ret)	/* add command clean_switch_mark */
-{
-    struct command *cmd;
-
-    cmd = add_command (cCLEAN_SWITCH_MARK, NULL);
-    cmd->args = keep;
-    cmd->tag = ret;
-}
-
-
-void
-clean_switch_mark (struct command *cmd)	/* pop everything up to (and including) first switch_mark from stack */
-{
-    struct stackentry *t, *tt, *b, *bb, *s;
-    int keep, k, ret;
-
-    k = keep = cmd->args;
-    ret = cmd->tag;
-    s = stackhead->prev;
-    while (k && s != stackroot) {
-        k--;
-        s = s->prev;
-    }
-    t = s;
-    tt = s->next;
-    while (((ret && s->type != stRETADDCALL)
-            || (!ret && s->type != stSWITCH_MARK)) && s != stackroot) {
-        s = s->prev;
-    }
-
-    if (!ret) {
-        while (s != stackroot && s->prev!=stackroot && s->prev->prev!=stackroot && s->prev->prev->type==stSWITCH_MARK) {
-            s = s->prev->prev;
-        }
-    }
-
-    if (ret) {
-        bb = s;
-        b = s->next;
-    } else {
-        b = s;
-        bb = s->prev;
-    }
-
-
-    /* cut part between (and including both) b and t out of stack */
-    bb->next = tt;
-    tt->prev = bb;
-    /* insert cut-out part between stackhead and stackhead->prev */
-    stackhead->prev->next = b;
-    b->prev = stackhead->prev;
-    t->next = stackhead;
-    stackhead->prev = t;
-    if (keep) {
-        stackhead = tt->next;
-    } else {
-        stackhead = bb->next;
-    }
-}
-
-
-void
-mybreak (struct command *cmd)	/* find break_here statement */
-{
-    struct command *curr;
-    int loop_nesting = 0;
-    int switch_nesting = 0;
-
-    curr = cmd;
-    while (curr->type != cBREAK_HERE || loop_nesting || switch_nesting) {
-        if (curr->type == cBEGIN_LOOP_MARK) {
-            loop_nesting++;
-        }
-        if (curr->type == cEND_LOOP_MARK) {
-            loop_nesting--;
-        }
-        if (curr->type == cBEGIN_SWITCH_MARK) {
-            switch_nesting++;
-        }
-        if (curr->type == cEND_SWITCH_MARK) {
-            switch_nesting--;
-        }
-        curr = curr->next;
-        if (!curr) {
-            sprintf(string,"break has left program (loop_nesting=%d, switch_nesting=%d)",loop_nesting,switch_nesting);
-            error (FATAL, string);
-        }
-    }
-    cmd->type = cQGOTO;
-    if (infolevel >= DEBUG) {
-        error (DEBUG, "converting cBREAK to cQGOTO");
-    }
-    cmd->jump = current = curr;
-}
-
-
-void
-mycontinue (struct command *cmd)	/* find continue_here statement */
-{
-    struct command *curr;
-    int loop_nesting = 0;
-
-    curr = cmd;
-
-    while (curr->type != cCONTINUE_HERE || loop_nesting) {
-        if (curr->type == cBEGIN_LOOP_MARK) {
-            loop_nesting++;
-        }
-        if (curr->type == cEND_LOOP_MARK) {
-            loop_nesting--;
-        }
-        curr = curr->prev;
-        if (!curr) {
-            sprintf(string,"continue has left program (loop_nesting=%d)",loop_nesting);
-            error (FATAL, string);
-        }
-    }
-    cmd->type = cQGOTO;
-    if (infolevel >= DEBUG) {
-        error (DEBUG, "converting cCONTINUE to cQGOTO");
-    }
-    cmd->jump = current = curr;
-}
-
-
-void
-next_case (struct command *cmd)	/* find next_case_here statement */
-{
-    struct command *curr;
-    int loop_nesting = 0;
-    int switch_nesting = 0;
-
-    curr = cmd;
-    while (curr->type != cNEXT_CASE_HERE || loop_nesting || switch_nesting) {
-        if (curr->type == cBEGIN_LOOP_MARK) {
-            loop_nesting++;
-        }
-        if (curr->type == cEND_LOOP_MARK) {
-            loop_nesting--;
-        }
-        if (curr->type == cBEGIN_SWITCH_MARK) {
-            switch_nesting++;
-        }
-        if (curr->type == cEND_SWITCH_MARK) {
-            switch_nesting--;
-        }
-        curr = curr->next;
-        if (!curr) {
-            sprintf(string,"search for next case has left program (loop_nesting=%d, switch_nesting=%d)",loop_nesting,switch_nesting);
-            error (FATAL, string);
-        }
-    }
-    cmd->type = cQGOTO;
-    if (infolevel >= DEBUG) {
-        error (DEBUG, "converting cNEXT_CASE to cQGOTO");
-    }
-    cmd->jump = current = curr;
-}
-
-
-void
-findnop ()
-/* used for on_gosub, find trailing nop command */
-{
-    while (current->type != cNOP) {
-        current = current->next;	/* next label */
-    }
-}
-
-
-void
-forcheck (void)			/* check, if for-loop is done */
-{
-    double start, bound, step, val;
-
-    val = pop (stNUMBER)->value;
-    step = pop (stNUMBER)->value;
-    bound = pop (stNUMBER)->value;
-    start = stackhead->prev->value;
-    if ((val <= bound && val >= start && step >= 0)
-            || (val <= start && val >= bound && step <= 0)) {
-        stackhead->prev->value = 1.;
-    } else {
-        stackhead->prev->value = 0.;
-    }
-}
-
-
-void
-forincrement (void)		/* increment value on stack */
-{
-    /* expecting on stack: BOUND,STEP,VAL,stackhead
-      where for VAL=START to BOUND step STEP */
-    stackhead->prev->value += stackhead->prev->prev->value;
-}
-
-
-void
-startfor (void)			/* compute initial value of for-variable */
-{
-    struct stackentry *p;
-
-    p = push ();
-    p->value =
-        stackhead->prev->prev->prev->prev->value -
-        stackhead->prev->prev->value;
-    p->type = stNUMBER;
-
-    return;
-}
