@@ -1,7 +1,7 @@
 /*
 
     YABASIC ---  a simple Basic Interpreter
-    written by Marc Ihm 1995-2016
+    written by Marc Ihm 1995-2017
     more info at www.yabasic.de
 
     yabasic.h --- function prototypes and global variables
@@ -119,7 +119,7 @@ extern int stream_modes[];	/* modes for streams */
 extern int curinized;		/* true, if curses has been initialized */
 extern int badstream (int, int);	/* test for valid stream id */
 void myseek (struct command *);	/* reposition file pointer */
-void myswitch (int);		/* switch to specified stream */
+void mystream (int);		/* switch to specified stream */
 #ifdef WINDOWS
 extern HANDLE gotwinkey;		/* mutex to signal key reception */
 extern char conkeybuff[]; /* Key received from console */
@@ -188,6 +188,7 @@ extern int backpid;		/* pid of process waiting for redraw events */
 
 /* function.c */
 extern struct command *datapointer;	/* current location for read-command */
+extern char *last_inkey; /* last result of inkey */
 void switch_compare (void);	/* compare topmost values for switch statement */
 
 /* symbol.c */
@@ -197,7 +198,6 @@ extern void query_array (struct command *cmd);	/* query array */
 extern struct command *lastref;	/* last command in UDS referencing a symbol */
 extern struct command *firstref;	/* first command in UDS referencing a symbol */
 extern int labelcount;		/* count self-generated labels */
-extern int switchtick;         /* change in and out of switch-statements */
 
 /* flex.c */
 extern int include_stack_ptr;	/* Lex buffer for any imported file */
@@ -214,8 +214,6 @@ extern char *current_function;	/* name of currently parsed function */
 extern int yydebug;
 extern int missing_endif;
 extern int missing_endif_line;
-extern int in_loop;
-extern int in_switch;
 
 /*-------------------------- defs and undefs ------------------------*/
 
@@ -252,14 +250,14 @@ enum error {
     FATAL, ERROR, INFO, DUMP, WARNING, NOTE, DEBUG
 };
 
-enum endreasons {
+enum end_reasons {
     /* ways to end the program */
     erNONE, erERROR, erREQUEST, erEOF
 };
 
-enum streammodes {
+enum stream_modes {
     /* ways to access a stream */
-    smCLOSED = 0, smREAD = 1, smWRITE = 2, smPRINT = 4
+    stmCLOSED = 0, stmREAD = 1, stmWRITE = 2, stmPRINT = 4
 };
 
 enum functions {
@@ -296,7 +294,7 @@ enum cmd_type {
     /* type of command */
     cFIRST_COMMAND,		/* no command, just marks start of list */
 
-    cLABEL, cSUBLINK, cGOTO, cQGOTO, cGOSUB, cQGOSUB, cRETURN,	/* flow control */
+    cLABEL, cLINK_SUBR, cGOTO, cQGOTO, cGOSUB, cQGOSUB, cRETURN_FROM_GOSUB,	/* flow control */
     cEND, cEXIT, cBIND, cDECIDE, cSKIPPER, cNOP, cFINDNOP, cEXCEPTION,
     cANDSHORT,
     cORSHORT, cSKIPONCE, cRESETSKIPONCE, cCOMPILE, cEXECUTE, cEXECUTE2,
@@ -306,18 +304,16 @@ enum cmd_type {
     cSPLIT, cSPLIT2, cSPLITALT, cSPLITALT2,
     cSTARTFOR, cFORCHECK, cFORINCREMENT,	/* for for-loops */
 
-    cSWITCH_COMPARE, cNEXT_CASE, cNEXT_CASE_HERE, cBREAK, 	/* break-continue-switch */
-    cCONTINUE, cBREAK_HERE, cCONTINUE_HERE,
-    cPUSH_SWITCH_MARK, cCLEAN_SWITCH_MARK,
+    cSWITCH_COMPARE, cNEXT_CASE, cNEXT_CASE_HERE, cBREAK_MULTI, 	/* break-continue-switch */
+    cCONTINUE, cBREAK_HERE, cCONTINUE_HERE, cPOP_MULTI,
     cBEGIN_LOOP_MARK, cEND_LOOP_MARK, cBEGIN_SWITCH_MARK, cEND_SWITCH_MARK,
 
-
     cDBLADD, cDBLMIN, cDBLMUL, cDBLDIV, cDBLPOW,	/* double operations */
     cNEGATE, cPUSHDBLSYM, cPOP, cPOPDBLSYM, cPUSHDBL,
 
-    cREQUIRE, cPUSHFREE, cMAKELOCAL, cMAKESTATIC, cNUMPARAM,	/* functions and procedures */
-    cCALL, cQCALL, cPUSHSYMLIST, cPOPSYMLIST, cRET_FROM_FUN,
-    cUSER_FUNCTION, cRETVAL, cEND_FUNCTION,
+    cREQUIRE, cPUSHFREE, cMAKELOCAL, cMAKESTATIC, cCOUNT_PARAMS,	/* functions and procedures */
+    cCALL, cQCALL, cPUSHSYMLIST, cPOPSYMLIST, cRETURN_FROM_CALL,
+    cUSER_FUNCTION, cCHECK_RETURN_VALUE, cEND_FUNCTION,
     cFUNCTION_OR_ARRAY, cSTRINGFUNCTION_OR_ARRAY,
 
     cPOKE, cPOKEFILE, cSWAP, cDUPLICATE, cDOCU,	/* internals */
@@ -344,9 +340,8 @@ enum cmd_type {
 enum stackentries {
     /* different types of stackentries */
     stGOTO, stSTRING, stSTRINGARRAYREF, stNUMBER, stNUMBERARRAYREF, stLABEL,
-    stRETADD, stRETADDCALL, stFREE, stROOT,
+    stRET_ADDR, stRET_ADDR_CALL, stFREE, stROOT,
     stANY, stSTRING_OR_NUMBER, stSTRING_OR_NUMBER_ARRAYREF,	/* these will never appear on stack but are used to check with pop */
-    stSWITCH_MARK,		/* used to clean up stack after switch-statement */
     stSWITCH_STRING, stSWITCH_NUMBER	/* only used in switch statement, compares true to every string or number */
 };
 
@@ -378,9 +373,9 @@ enum yabkeys {
     kBACKSPACE, kSCRNDOWN, kSCRNUP, kENTER, kESC, kTAB, kLASTKEY
 };
 
-enum searchmodes {
+enum search_modes {
     /* modes for searching labels */
-    smSUB = 1, smLINK = 2, smLABEL = 4, smGLOBAL = 8
+    srmSUBR = 1, srmLINK = 2, srmLABEL = 4, srmGLOBAL = 8
 };
 
 /* ------------- global types ---------------- */
@@ -429,7 +424,7 @@ struct command {
     void *pointer;		/* pointer to data */
     void *symbol;			/* pointer to symbol (or data within symbol) associated with command */
     struct command *jump;		/* pointer to jump destination */
-    char *name;			/* name of symbol associated with command */
+    char *symname;			/* name of symbol associated with command */
     struct command *nextref;	/* next cmd within function referencing a symbol */
     struct command *nextassoc;	/* next cmd within within chain of associated commands */
     int args;			/* number of arguments for function/array call */
@@ -437,9 +432,15 @@ struct command {
     int tag;			/* char/int to pass some information */
     int line;			/* line this command has been created for */
     struct libfile_name *lib;	/* associated library */
-    int switchtick;		/* TRUE, if in switch, FALSE else; used to check gotos */
+    char *diag;                 /* optional text for diagnostics */
+    struct switch_state *switch_state;  /* state for switch statements */
 };
 
+struct switch_state {    /* records surrounding of a statement; used to check gotos into, out-of, within and between switch-statements (i.e. to err on most)  */
+    int id;         /* unique id for each switch-statment */
+    int nesting;    /* number of nested switch-statements */
+};
+    
 struct array {
     /* data structure for arrays */
     int bounds[10];		/* index boundaries */
@@ -472,11 +473,12 @@ struct libfile_name {
 /* main.c */
 void error (int, char *);	/* reports an error and possibly exits */
 void error_with_line (int, char *, int);	/* reports an error and possibly exits */
-void std_diag (char *, int, char *);	/* produce standard diagnostic */
+void std_diag (char *, int, char *, char *);	/* produce standard diagnostic */
 void *my_malloc (unsigned);	/* my own version of malloc */
 void my_free (void *);		/* free memory */
 char *my_strerror (int);	/* return description of error messages */
-struct command *add_command (int, char *);	/* get room for new command */
+struct command *add_command (int, char *, char *);	/* create new command and add some data */
+struct command *add_command_with_switch_state(int);     /* same as add_command, but add switch_state too */
 void dump_commands (char *);         /* dump commands into given file */
 void signal_handler (int);	/* handle various signals */
 char *my_strdup (char *);	/* my own version of strdup */
@@ -495,8 +497,8 @@ int isbound (void);		/* check if this interpreter is bound to a program */
 
 /* io.c */
 void checkopen (void);		/* check, if open has been sucessfull */
-void create_colour (int);	/* create command 'reverse' */
-void colour (struct command *cmd);	/* switch reverse-printing */
+void create_colour (int);       /* create command 'colour' */
+void colour (struct command *cmd);	/* change colour of printed output */
 void create_print (char);	/* create command 'print' */
 void print (struct command *);	/* print on screen */
 void create_myread (char, int);	/* create command 'read' */
@@ -508,9 +510,9 @@ void create_myopen (int);	/* create command 'myopen' */
 void myopen (struct command *);	/* open specified file for given name */
 void testeof (struct command *);	/* close the specified stream */
 void myclose ();		/* close the specified stream */
-void create_pps (int, int);	/* create command pushswitch or popswitch */
-void push_switch (struct command *);	/* push current stream on stack and switch to new one */
-void pop_switch (void);		/* pop current stream from stack and switch to it */
+void create_pps (int, int);	/* create command push_stream or pop_stream */
+void push_stream (struct command *);	/* push current stream on stack and switch to new one */
+void pop_stream (void);		/* pop stream from stack and switch to it */
 void mymove ();			/* move to specific position on screen */
 void clearscreen ();		/* clear entire screen */
 char *inkey (double);		/* gets char from keyboard, blocks and doesnt print */
@@ -597,8 +599,6 @@ void pusharrayref (struct command *);	/* push an array reference onto stack */
 void arraylink (struct command *);	/* link a local symbol to a global array */
 void makestatic (struct command *);	/* makes symbol static */
 void makelocal (struct command *);	/* makes symbol local */
-void create_numparam (void);	/* create command 'cNUMPARAM' */
-void numparam (struct command *);	/* count number of function parameters */
 void pushdblsym (struct command *);	/* push double symbol onto stack */
 void popdblsym (struct command *);	/* pop double from stack */
 void create_pushdbl (double);	/* create command 'pushdbl' */
@@ -611,6 +611,29 @@ void create_pushstr (char *);	/* creates command pushstr */
 void pushstr (struct command *);	/* push string onto stack */
 void pushname (char *);		/* push a name on stack */
 void pushstrptr (struct command *);	/* push string-pointer onto stack */
+void logical_shortcut (struct command *type);	/* shortcut and/or if possible */
+void create_doarray (char *, int);	/* creates array-commands */
+void doarray (struct command *);	/* call an array */
+void create_dim (char *, char);	/* create command 'dim' */
+void dim (struct command *);	/* get room for array */
+void swap (void);		/*swap topmost elements on stack */
+struct stackentry *push (void);	/* push element on stack and enlarge it */
+struct stackentry *pop (int);	/* pops element to memory */
+void pop_multi (struct command *); /* pop and discard multiple values from stack */
+struct symbol *get_sym (char *, int, int);	/* find and/or add a symbol */
+void link_symbols (struct symbol *, struct symbol *);	/* link one symbol to the other */
+void pushsymlist (void);	/* push a new list on symbol stack */
+void popsymlist (void);		/* pop list of symbols and free symbol contents */
+void dump_sym ();		/* dump the stack of lists of symbols */
+void dump_sub (int);		/* dump the stack of subroutine calls */
+void function_or_array (struct command *);	/* decide whether to do perform function or array */
+int count_args (int);	/* count number of arguments on stack */
+
+
+/* flow.c */
+void link_label (struct command *);	/* link label into list of labels */
+void create_count_params (void);	/* create command 'cCOUNT_PARAMS' */
+void count_params (struct command *);	/* count number of function parameters */
 void forcheck (void);		/* check, if for-loop is done */
 void forincrement (void);	/* increment value on stack */
 void startfor (void);		/* compute initial value of for-variable */
@@ -618,7 +641,8 @@ void create_goto (char *);	/* creates command goto */
 void create_gosub (char *);	/* creates command gosub */
 void create_call (char *);	/* creates command function call */
 void create_label (char *, int);	/* creates command label */
-void create_sublink (char *);	/* create link to subroutine */
+void create_subr_link (char *);	/* create link to subroutine */
+struct command *add_switch_state(struct command *); /* add switch state to a newly created command */
 void pushgoto (void);		/* generate label and push goto on stack */
 void popgoto (void);		/* pops a goto and generates the matching command */
 void jump (struct command *);	/* jump to specific Label */
@@ -628,36 +652,26 @@ void skipper (void);		/* used for on_goto/gosub, skip commands */
 void skiponce (struct command *);	/* skip next command once */
 void resetskiponce (struct command *);	/* find and reset next skip */
 void decide (void);		/*  skips next command, if not 0 on stack */
-void logical_shortcut (struct command *type);	/* shortcut and/or if possible */
-void create_doarray (char *, int);	/* creates array-commands */
-void doarray (struct command *);	/* call an array */
-void create_dim (char *, char);	/* create command 'dim' */
-void dim (struct command *);	/* get room for array */
 void pushlabel (void);		/* generate goto and push label on stack */
 void poplabel (void);		/* pops a label and generates the matching command */
 void storelabel ();		/* push label on stack */
 void matchgoto ();		/* generate goto matching label on stack */
-void swap (void);		/*swap topmost elements on stack */
-struct stackentry *push (void);	/* push element on stack and enlarge it */
-struct stackentry *pop (int);	/* pops element to memory */
-struct symbol *get_sym (char *, int, int);	/* find and/or add a symbol */
-void link_symbols (struct symbol *, struct symbol *);	/* link one symbol to the other */
-void pushsymlist (void);	/* push a new list on symbol stack */
-void popsymlist (void);		/* pop list of symbols and free symbol contents */
-void dump_sym ();		/* dump the stack of lists of symbols */
-void dump_sub (int);		/* dump the stack of subroutine calls */
-void create_retval (int, int);	/* create command 'cRETVAL' */
-void retval (struct command *);	/* check return value of function */
+void create_check_return_value (int, int);	/* create command 'cCHECK_RETURN_VALUE' */
+void check_return_value (struct command *);	/* check return value of function */
 void create_endfunction (void);	/* create command cEND_FUNCTION */
-void function_or_array (struct command *);	/* decide whether to do perform function or array */
 struct command *search_label (char *, int);	/* search label */
-void reshufflestack (struct stackentry *);	/* reorganize stack for function call */
-void push_switch_mark (void);	/* push a switch mark */
-void create_clean_switch_mark (int, int);	/* add command clean_switch_mark */
-void clean_switch_mark (struct command *);	/* pop everything up to (and including) first switch_mark from stack */
+void reorder_stack_before_call (struct stackentry *);	/* reorganize stack for function call */
+void reorder_stack_after_call (int);	/* reorganize stack after function call: keep return value and remove switch value (if any) */
+void create_mybreak(int); /* create command mybreak */
 void mybreak (struct command *);	/* find break_here statement */
 void mycontinue (struct command *cmd);	/* find continue_here statement */
 void next_case (struct command *);		/* go to next case in switch statement */
+void check_leave_switch (struct command *, struct command *);    /* check, if goto or continue enters or leaves a switch_statement */
+void pop_switch_value (struct command *); /* remove switch state from stack, keeping return value */
+void initialize_switch_id_stack(void); /* initialize stack of switch_ids */
+void push_switch_id (void);		  /* generate a new switch id on top of stack */
+void pop_switch_id (void);  		  /* pop last switch_id */
+
 
 /* flex.c */
 void yyerror (char *);		/* yyerror message */
